#ifndef HUFFMAN_H
#define HUFFMAN_H
/*
 * Copyright (c) Alex Holden <alex@linuxhacker.org> 2000.
 *
 * This code is public domain; you can do whatever you want with it, though I
 * would prefer you to contribute any bug fixes back to me if possible.
 * This software comes with NO WARRANTY WHATSOEVER, not even the implied
 * warranties of merchantability and fitness for a particular purpose.
 */

#include "heap.h"

/*
 * The minimum length before we should even attempt to compress a block-
 * the reason this is so high is because of the 256 byte header which is
 * needed to generate the decoding tree and the fact that the absolute
 * minimum length a byte can compress down to is one bit, but the output data
 * is always an exact number of bytes. So the minimum length a block has to be
 * before it stands any chance of compressing to less than it's uncompressed
 * size is 256 + 256/8 + 2 = 290. Arguably it probably isn't really worth
 * wasting processing time on blocks which are only slightly higher than this
 * either due to the fact that they can't possibly compress by very much.
 */
#define HUFF_MIN_ENC 290

typedef struct huff_node {
	int freq;		/* Occurence frequency */
	struct huff_node *lc;	/* Left hand child node */
	struct huff_node *rc;	/* Right hand child node */
	u16 code;		/* Huffman code */
	u8 length;		/* Length of code */
	u16 value;		/* The character associated with this node */
} huffnode;

typedef struct {
	int freq;		/* The combined frequencies of the children */
	huffnode *lc;		/* Left hand child node */
	huffnode *rc;		/* Right hand child node */
} huffnodule;

typedef struct {
	int buflen;		/* The length of the output buffer */
	u8 *inbuf;		/* Pointer to the input buffer */
	u8 *outbuf;		/* Pointer to the output buffer */
	huffnode nodes[257];	/* Array of huffman tree nodes */
	int nodule;		/* Current nodule number */
	huffnodule nodules[256];/* Array of nodules */
	huffnode *root;		/* Root of huffman binary tree */
	heap *hp;		/* Pointer to heap structure */
} huffstate;

extern huffstate *huff_init(int buflen);
extern void huff_destroy(huffstate *state);
extern int huffenc(huffstate *state, int);
extern int huffdec(huffstate *state, int);

#ifdef HUFFMAN_PRIV

static void make_scaled_freqs(huffstate *state, int len);
static void make_huff_tree(huffstate *state);
static void makehuffenctab(huffstate *state, huffnode *node, u16 code, u8 len);
#define make_huff_enc_table(A) makehuffenctab((A), (A)->root, 0, 0)
static int compare_huffnodes(huffnode *a, huffnode *b);
static void cleantrees(huffstate *state);

#endif
#endif
