/*
 * TEA Total Copyright (c) Alex Holden <alex@linuxhacker.org> 2000, 2001.
 *
 * This code is public domain; you can do whatever you want with it, though I
 * would prefer you to contribute any bug fixes back to me if possible.
 * This software comes with NO WARRANTY WHATSOEVER, not even the implied
 * warranties of merchantability and fitness for a particular purpose.
 *
 * teatotal.c: The main file of the teatotal program.
 */

#include <string.h>
#include <unistd.h>

#include "teatotal.h"

#ifdef BUILD_TEA_UNTEA
#include "teauntea.h"
#endif

#ifdef BUILD_TEA_KGEN
#include "tea-kgen.h"
#endif

/*
 * Print the usage message for the program that we were started as and exit.
 */
void usage(teastate *state)
{
	switch(state->program) {
		case PROGRAM_TEATOTAL:
			die("Usage: teatotal <applet-name> [applet-options]"
#ifdef DOS_LINE_ENDINGS
				"\r\n"
#else
				"\n"
#endif
				"applets:"
#ifdef BUILD_TEA_UNTEA
				" tea untea"
#endif
#ifdef BUILD_TEA_KGEN
				" tea-kgen"
#endif
				);
#ifdef BUILD_TEA_UNTEA
		case PROGRAM_TEA:
			die("Usage: tea "
#ifdef BUILD_BASE64
				"[ -a ] "
#endif
#ifdef BUILD_HUFFMAN
				"[ -c ] "
#endif
#ifdef BUILD_PPKEYS
				"[ -p pass ] "
#endif
				"[ -h ] <key-file> [ infile [ outfile ] ]");
		case PROGRAM_UNTEA:
			die("Usage: untea "
#ifdef BUILD_BASE64
				"[ -a ] "
#endif
#ifdef BUILD_PPKEYS
				"[ -p pass ] "
#endif
				"<key-file> [ infile [ outfile ] ]");
#endif /* BUILD_TEA_UNTEA */
#ifdef BUILD_TEA_KGEN
		case PROGRAM_TEA_KGEN:
			die("Usage: tea-kgen "
#ifdef BUILD_PPKEYS
				"[ -c ] [ -P ] [ -p pass ] [ -k pass ] "
#endif
				"[ -h ] <key-file>");
#endif
		case PROGRAM_UNKNOWN:
		default:
			die("Invalid applet name");
	}
}

/*
 * Check for unused command line arguments, and if VERBOSE_ARGUMENT_ERRORS
 * is defined, print an error message about each one.
 */
static void check_bad_args(teastate *state)
{
	int args_ok = 1;

#ifdef VERBOSE_ARGUMENT_ERRORS
	char *p, c;

	/* Check for invalid arguments */
	while((p = getarg(state->ga, 0, NULL))) {
		safe_write(STDERR_FILENO, "Invalid argument: ", 18);
		safe_write(STDERR_FILENO, p, strlen(p));
#ifdef DOS_LINE_ENDINGS
		safe_write(STDERR_FILENO, "\r\n", 2);
#else
		safe_write(STDERR_FILENO, "\n", 1);
#endif
		args_ok = 0;
	}
	while((c = getflag(state->ga, 0, NULL))) {
		safe_write(STDERR_FILENO, "Invalid short flag: -", 21);
		safe_write(STDERR_FILENO, &c, 1);
#ifdef DOS_LINE_ENDINGS
		safe_write(STDERR_FILENO, "\r\n", 2);
#else
		safe_write(STDERR_FILENO, "\n", 1);
#endif
		args_ok = 0;
	}
	while((p = getarg(state->ga, 0, ""))) {
		safe_write(STDERR_FILENO, "Invalid long flag: ", 20);
		safe_write(STDERR_FILENO, p, strlen(p));
#ifdef DOS_LINE_ENDINGS
		safe_write(STDERR_FILENO, "\r\n", 2);
#else
		safe_write(STDERR_FILENO, "\n", 1);
#endif
		args_ok = 0;
	}
#else
	if(getarg(state->ga, 0, NULL) || getarg(state->ga, 0, "") ||
					getflag(state->ga, 0, NULL))
		args_ok = 0;
#endif

	if(!args_ok) usage(state);
}

/*
 * Initialise the program.
 */
void initialise(teastate *state, int argc, char *argv[])
{
	char *p = argv[0];
	char *appname = p;

	/* Strip leading path from the name */
	while(*p) if(*p++ == '/') appname = p;

	/* Check if we were called directly as "teatotal"; if so, take the
	 * first argument to be the program name instead */
	if(!strcmp(appname, "teatotal")) {
		if(argc < 2) {
			state->program = PROGRAM_TEATOTAL;
			usage(state);
		}
		argc--;
		appname = argv[1];
		argv = &argv[1];
	}

	/* Initialise the argument parser */
	if(!(state->ga = getarg_init(argc, argv)))
		die("Out of memory");

#ifdef BUILD_PPKEYS
	/* Set the key passwords off by default */
	state->inkeypass = NULL;
	state->outkeypass = NULL;
	state->keyfile = NULL;
#endif

	/* Find out what applet we were called as and call the relevant init()*/
	if(0) {
#ifdef BUILD_TEA_UNTEA
	} else if(!strcmp(appname, "tea")) {
		state->program = PROGRAM_TEA;
		state->untea = 0;
		init_tea_untea(state);
	} else if(!strcmp(appname, "untea")) {
		state->program = PROGRAM_UNTEA;
		state->untea = 1;
		init_tea_untea(state);
#endif /* BUILD_TEA_UNTEA */
#ifdef BUILD_TEA_KGEN
	} else if(!strcmp(appname, "tea-kgen")) {
		state->program = PROGRAM_TEA_KGEN;
		init_tea_kgen(state);
#endif /* BUILD_TEA_KGEN */
	} else {
		state->program = PROGRAM_UNKNOWN;
		usage(state);
	}

	/* Check that all of the command line arguments were used */
	check_bad_args(state);

	/* Free the argument parser structure */
	getarg_destroy(state->ga);
}

int main(int argc, char *argv[])
{
	teastate *state;

	/* Allocate the global state structure */
	state = safe_malloc(sizeof(teastate));

	/* Initialise the program */
	initialise(state, argc, argv);

	/* Execute the relevant applet */
	switch(state->program) {
#ifdef BUILD_TEA_UNTEA
		case PROGRAM_TEA:
		case PROGRAM_UNTEA:
			do_tea_untea(state);
			break;
#endif
#ifdef BUILD_TEA_KGEN
		case PROGRAM_TEA_KGEN:
			do_tea_kgen(state);
			break;
#endif
	}

	return 0;
}
